"""
Taken from
https://github.com/allenai/allennlp-models/blob/main/allennlp_models/version.py
"""
_MAJOR = "0"
_MINOR = "0"
_PATCH = "7"

VERSION_SHORT = f"{_MAJOR}.{_MINOR}"
VERSION = f"{VERSION_SHORT}.{_PATCH}"
__version__ = VERSION
